# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import xml.dom.minidom
import re

from typing import List, Iterable, Optional, Tuple, Union, TYPE_CHECKING

from .. import utils
from .. import scoring
from . import questions
from . import parametric
from . import words

if TYPE_CHECKING:
    import xml.dom
    import xml.dom.minicompat


EyegradeException = utils.EyegradeException

EYEGRADE_NAMESPACE = "http://www.it.uc3m.es/jaf/eyegrade/ns/"
text_norm_re = re.compile(r"[\ \t\n]+")

# Register user-friendly error messages
EyegradeException.register_error(
    "exam_root_element",
    "The root element of the exam XML with the questions is expected\n"
    "to be named 'exam' in the Eyegrade namespace. Review the\n"
    "syntax of your exam and namespace declarations.\n"
    "The Eyegrade namespace is: " + EYEGRADE_NAMESPACE,
)
EyegradeException.register_error(
    "exam_one_choices",
    short_message="Only one 'choices' element is allowed in a question.",
)
EyegradeException.register_error(
    "bad_position_value",
    "The value of 'position' must be either 'center' or 'right'.",
    "Incorrect value for the 'position' attribute.",
)
EyegradeException.register_error(
    "missing_width_code",
    "Code positioned to the right must have a 'width' attribute.",
    "Missing 'width' attribute in a 'code' element.",
)
EyegradeException.register_error(
    "missing_width_fig",
    "Figures must always have a 'width' attribute.",
    "Missing 'width' attribute in a 'figure' element.",
)
EyegradeException.register_error(
    "missing_text",
    "Questions must contain exactly one 'text' element.",
    "Text element expected in a question.",
)
EyegradeException.register_error(
    "duplicate_text",
    "Questions must contain exactly one 'text' element, and groups one 'common' element.",
    "Multiple text elements in a question or common elements in a group.",
)
EyegradeException.register_error(
    "score_correct_needed",
    "When a score for incorrect answers is provided, "
    "a score for correct answers is also needed.",
)
EyegradeException.register_error(
    "duplicate_studentid_element", "At most one studentId element can be provided."
)
EyegradeException.register_error(
    "studentid_number_format", "Student id length must be an integer"
)
EyegradeException.register_error(
    "duplicate_score_element", "At most one score element can be provided."
)
EyegradeException.register_error(
    "duplicate_points_words_element", "At most one pointsWords element can be provided."
)
EyegradeException.register_error(
    "empty_score_element",
    "The score element must specify either a maximum score "
    "or a score for correct answers.",
)
EyegradeException.register_error(
    "incorrect_score_element",
    "The score element must specify either a maximum score "
    "or scores for correct and incorrect answers, but not both.",
)
EyegradeException.register_error(
    "penalize_attribute",
    "The value of the penalize attribute must be either 'true' or 'false'.",
)
EyegradeException.register_error(
    "incompatible_variation",
    "Variations of the same question must contain "
    "the same number of correct and incorrect choices. "
    "Fixed choices must appear at the same positions.",
)
EyegradeException.register_error(
    "incompatible_variation_num_group",
    "Questions within the same group must contain the same number of variations.",
)
EyegradeException.register_error(
    "incompatible_variations_declarations",
    "Incompatible elements within the same question: <variation> and <variation_params>",
)
EyegradeException.register_error(
    "multiple_variation_params",
    "At most one <variation_params> element per question is allowed",
)
EyegradeException.register_error(
    "missing_param_name", "Missing 'eye:name' attribute in <param> element"
)
EyegradeException.register_error("bad_fix_value", "Bad value for eye:fix attribute")


def parse_exam(exam_filename: str) -> questions.ExamQuestions:
    """Parses the questions of a exam from an XML file."""
    dom_tree = xml.dom.minidom.parse(exam_filename)
    # By now, only one parser exists. In the future multiple parsers can
    # be called from here, to allow multiple data formats.
    return _parse_tree(dom_tree)


def _parse_tree(dom_tree: xml.dom.minidom.Document) -> questions.ExamQuestions:
    assert dom_tree.nodeType == xml.dom.minidom.Node.DOCUMENT_NODE
    root = dom_tree.childNodes[0]
    if isinstance(root, xml.dom.minidom.Element) and get_full_name(root) == (
        EYEGRADE_NAMESPACE,
        "exam",
    ):
        exam = questions.ExamQuestions()
        exam.subject = get_element_content(root, EYEGRADE_NAMESPACE, "subject")
        exam.degree = get_element_content(root, EYEGRADE_NAMESPACE, "degree")
        exam.date = get_element_content(root, EYEGRADE_NAMESPACE, "date")
        exam.duration = get_element_content(root, EYEGRADE_NAMESPACE, "duration")
        exam.title = get_element_content(root, EYEGRADE_NAMESPACE, "title")
        student_id_length, student_id_label = parse_student_id(root)
        if student_id_length is not None:
            exam.student_id_length = student_id_length
        if student_id_label is not None:
            exam.student_id_label = student_id_label
        for node in get_children_by_tag_names(
            root, EYEGRADE_NAMESPACE, ["question", "group"]
        ):
            if node.localName == "question":
                exam.questions.append(parse_question(node))
            else:
                exam.questions.append(parse_group(node))
        scores = parse_scores(root)
        if isinstance(scores, scoring.AutomaticScore):
            exam.scores = scores.compute(exam.num_questions(), exam.num_choices())
        else:
            exam.scores = scores
        exam.points_words = parse_points_words(root)
    else:
        raise EyegradeException("Bad root element", key="exam_root_element")
    return exam


def parse_student_id(
    root: xml.dom.minidom.Element,
) -> Tuple[Optional[int], Optional[str]]:
    student_id_length = None
    student_id_label = None
    element_list = get_children_by_tag_name(root, EYEGRADE_NAMESPACE, "studentId")
    if len(element_list) == 1:
        student_id_element = element_list[0]
        student_id_length_str = get_attribute_text(student_id_element, "length")
        student_id_label = get_attribute_text(student_id_element, "label")
        if student_id_length_str is not None:
            try:
                student_id_length = int(student_id_length_str)
            except ValueError:
                raise EyegradeException("", key="student_id_number_format")
    elif len(element_list) > 1:
        raise EyegradeException("", key="duplicate_studentid_element")
    return student_id_length, student_id_label


def parse_scores(
    root: xml.dom.minidom.Element,
) -> Optional[Union[scoring.AutomaticScore, scoring.QuestionScores]]:
    scores: Optional[Union[scoring.AutomaticScore, scoring.QuestionScores]] = None
    element_list = get_children_by_tag_name(root, EYEGRADE_NAMESPACE, "scores")
    if len(element_list) == 1:
        score_element = element_list[0]
        max_score_attr = get_attribute_text(score_element, "maxScore")
        penalize_attr = get_attribute_text(score_element, "penalize")
        correct_attr = get_attribute_text(score_element, "correct")
        incorrect_attr = get_attribute_text(score_element, "incorrect")
        if max_score_attr is not None:
            # Automatically compute scores from the maximum score
            if correct_attr is not None or incorrect_attr is not None:
                raise EyegradeException("", key="incorrect_score_element")
            if penalize_attr == "true":
                penalize = True
            elif penalize_attr is None or penalize_attr == "false":
                penalize = False
            else:
                raise EyegradeException("", key="penalize_attribute")
            scores = scoring.AutomaticScore(max_score_attr, penalize)
        elif correct_attr is not None:
            if incorrect_attr is None:
                incorrect_attr = "0"
            scores = scoring.QuestionScores(correct_attr, incorrect_attr, "0")
        elif incorrect_attr is not None:
            raise EyegradeException("", key="score_correct_needed")
        else:
            raise EyegradeException("", key="empty_score_element")
    elif len(element_list) > 1:
        raise EyegradeException("", key="duplicate_score_element")
    return scores


def parse_points_words(root: xml.dom.minidom.Element) -> Optional[words.PointsWords]:
    element_list = get_children_by_tag_name(root, EYEGRADE_NAMESPACE, "pointsWords")
    if len(element_list) == 1:
        element = element_list[0]
        singular_word = get_attribute_text(element, "singular")
        plural_word = get_attribute_text(element, "plural")
        if singular_word is not None and plural_word is not None:
            pointsWords = words.PointsWords(singular_word, plural_word)
        elif singular_word is not None:
            pointsWords = words.PointsWords(singular_word)
        elif plural_word is not None:
            pointsWords = words.PointsWords(plural_word)
        else:
            pointsWords = None
    elif not element_list:
        pointsWords = None
    else:
        raise EyegradeException("", key="duplicate_points_words_element")
    return pointsWords


def parse_question(question_node: xml.dom.minidom.Element) -> questions.Question:
    question: questions.Question
    variation_nodes = get_children_by_tag_name(
        question_node, EYEGRADE_NAMESPACE, "variation"
    )
    parameter_sets = parse_parameter_sets(question_node)
    if variation_nodes and parameter_sets:
        raise EyegradeException("", key="incompatible_variations_declarations")
    if variation_nodes:
        question = questions.Question()
        for node in variation_nodes:
            question.add_variation(parse_question_variation(node))
    elif parameter_sets:
        question = parse_parametric_question(question_node, parameter_sets)
    else:
        question = questions.FixedQuestion(parse_question_variation(question_node))
    return question


def parse_question_variation(
    variation_node: xml.dom.minidom.Element,
) -> questions.QuestionVariation:
    text = parse_question_component(variation_node, False)
    choices_list = get_children_by_tag_name(
        variation_node, EYEGRADE_NAMESPACE, "choices"
    )
    if len(choices_list) != 1:
        raise EyegradeException("", key="exam_one_choices")
    choices = choices_list[0]
    correct_choices = []
    incorrect_choices = []
    fix_first: List[int] = []
    fix_last: List[int] = []
    counter = 0
    for node in get_children_by_tag_name(choices, EYEGRADE_NAMESPACE, "correct"):
        correct_choices.append(parse_question_component(node, True))
        read_fix_attr(node, fix_first, fix_last, counter)
        counter += 1
    for node in get_children_by_tag_name(choices, EYEGRADE_NAMESPACE, "incorrect"):
        incorrect_choices.append(parse_question_component(node, True))
        read_fix_attr(node, fix_first, fix_last, counter)
        counter += 1
    tags = get_element_content(variation_node, EYEGRADE_NAMESPACE, "tags")
    return questions.QuestionVariation(
        text, correct_choices, incorrect_choices, fix_first, fix_last, tags=tags
    )


def read_fix_attr(
    node: xml.dom.minidom.Element, fix_first: List[int], fix_last: List[int], index: int
) -> None:
    value = get_attribute_text(node, "fix")
    if not value:
        pass
    elif value == "first":
        fix_first.append(index)
    elif value == "last":
        fix_last.append(index)
    else:
        raise EyegradeException(
            f"Got '{value}' but expected 'first' or 'last'", key="bad_fix_value"
        )


def parse_parametric_question(
    node: xml.dom.minidom.Element, parameter_sets: List[parametric.ParameterSet]
) -> parametric.ParametricQuestion:
    question = parametric.ParametricQuestion(parse_question_variation(node))
    for parameter_set in parameter_sets:
        question.add_parameter_set(parameter_set)
    return question


def parse_parameter_sets(
    parent: xml.dom.minidom.Element,
) -> List[parametric.ParameterSet]:
    variation_params_nodes = get_children_by_tag_name(
        parent, EYEGRADE_NAMESPACE, "variation_params"
    )
    if variation_params_nodes:
        if len(variation_params_nodes) > 1:
            raise EyegradeException("", key="multiple_variation_params")
        return parse_variation_params_node(variation_params_nodes[0])
    return []


def parse_group(group_node: xml.dom.minidom.Element) -> questions.QuestionsGroup:
    question_list: List[questions.Question] = []
    common_text: Optional[questions.GroupCommonComponent]
    parameter_sets = parse_parameter_sets(group_node)
    common_text = _parse_group_common(group_node, parameter_sets)
    for node in get_children_by_tag_name(group_node, EYEGRADE_NAMESPACE, "question"):
        if not parameter_sets:
            question = parse_question(node)
        else:
            question = parse_parametric_question(node, parameter_sets)
        question_list.append(question)
    return questions.QuestionsGroup(question_list, common_text=common_text)


def _parse_group_common(
    group_node: xml.dom.minidom.Element, parameter_sets: List[parametric.ParameterSet]
) -> Optional[questions.GroupCommonComponent]:
    common_text: Optional[questions.GroupCommonComponent] = None
    element_list = get_children_by_tag_name(group_node, EYEGRADE_NAMESPACE, "common")
    if len(element_list) == 1:
        common_node = element_list[0]
        variation_nodes = get_children_by_tag_name(
            common_node, EYEGRADE_NAMESPACE, "variation"
        )
        if variation_nodes and parameter_sets:
            raise EyegradeException("", key="incompatible_variations_declarations")
        if variation_nodes:
            common_text = questions.GroupCommonComponent()
            for node in variation_nodes:
                common_text.add_variation(parse_question_component(node, False))
        elif parameter_sets:
            common_text = parametric.ParametricGroupCommonComponent(
                parse_question_component(common_node, False)
            )
            for parameter_set in parameter_sets:
                common_text.add_parameter_set(parameter_set)
        else:
            common_text = questions.FixedGroupCommonComponent(
                parse_question_component(common_node, False)
            )
    elif len(element_list) > 1:
        raise EyegradeException("", key="duplicate_text")
    return common_text


def parse_question_component(
    parent_node: xml.dom.minidom.Element, is_choice: bool
) -> questions.QuestionComponent:
    component = questions.QuestionComponent(is_choice)
    if not is_choice:
        component.text = get_question_text_content(parent_node, EYEGRADE_NAMESPACE)
    else:
        component.text = get_element_content_node(parent_node)
    component.code, code_attrs = get_element_content_with_attrs(
        parent_node, EYEGRADE_NAMESPACE, "code", ["width", "position", "separation"]
    )
    component.figure, figure_attrs = get_element_content_with_attrs(
        parent_node, EYEGRADE_NAMESPACE, "figure", ["width", "position", "separation"]
    )
    if component.code is not None:
        if code_attrs[1] is None:
            code_attrs[1] = "center"
        elif code_attrs[1] != "center" and code_attrs[1] != "right":
            raise EyegradeException("", key="bad_position_value")
        if code_attrs[0] is None and code_attrs[1] == "right":
            raise EyegradeException("", key="missing_width_code")
        if code_attrs[0] is not None:
            component.annex_width = float(code_attrs[0])
        else:
            component.annex_width = None
        component.annex_pos = code_attrs[1]
        if code_attrs[2] and code_attrs[2] == "line":
            component.annex_separation = True
        else:
            component.annex_separation = False
    if component.figure is not None:
        if figure_attrs[1] is None:
            figure_attrs[1] = "center"
        elif figure_attrs[1] != "center" and figure_attrs[1] != "right":
            raise EyegradeException("", key="bad_position_value")
        if figure_attrs[0] is None:
            raise EyegradeException("", key="missing_width_fig")
        component.annex_width = float(figure_attrs[0])
        component.annex_pos = figure_attrs[1]
        if figure_attrs[2] and figure_attrs[2] == "line":
            component.annex_separation = True
        else:
            component.annex_separation = False
    component.check_is_valid()
    return component


def parse_variation_params_node(
    node: xml.dom.minidom.Element,
) -> List[parametric.ParameterSet]:
    parameter_sets = []
    for variation_node in get_children_by_tag_name(
        node, EYEGRADE_NAMESPACE, "variation"
    ):
        parameter_sets.append(parse_parameter_set(variation_node))
    return parameter_sets


def parse_parameter_set(node: xml.dom.minidom.Element) -> parametric.ParameterSet:
    parameter_set = parametric.ParameterSet()
    for parameter_node in get_children_by_tag_name(node, EYEGRADE_NAMESPACE, "param"):
        name = get_attribute_text(parameter_node, "name")
        if name is None:
            raise EyegradeException("", key="missing_param_name")
        value = get_element_content_node(parameter_node)
        if value is None:
            value = ""
        parameter_set.add_parameter(name, value)
    return parameter_set


def get_question_text_content(
    parent: xml.dom.minidom.Element, namespace: str
) -> List[Tuple[str, str]]:
    parts = []
    node_list = get_children_by_tag_name(parent, namespace, "text")
    if len(node_list) == 1:
        for node in node_list[0].childNodes:
            if node.nodeType == node.TEXT_NODE:
                parts.append(("text", node.data.strip()))
            #                parts.append(('text', text_norm_re.sub(' ', node.data)))
            elif node.nodeType == node.ELEMENT_NODE:
                if node.namespaceURI == namespace and node.localName == "code":
                    text = get_text(node.childNodes, False)
                    parts.append(("code", text if text is not None else ""))
                else:
                    raise EyegradeException("Unknown element: " + node.localName)
    elif not node_list:
        raise EyegradeException("", key="missing_text")
    else:
        raise EyegradeException("", key="duplicate_text")
    return parts


def get_element_content(
    parent: xml.dom.minidom.Element, namespace: str, local_name: str
) -> Optional[str]:
    content: Optional[str]
    node_list = get_children_by_tag_name(parent, namespace, local_name)
    if not node_list:
        content = None
    elif len(node_list) == 1:
        content = get_text(node_list[0].childNodes)
    else:
        raise EyegradeException("Duplicate element: " + local_name)
    return content


def get_element_content_node(element_node: xml.dom.minidom.Element) -> Optional[str]:
    return get_text(element_node.childNodes, False)


def get_element_content_with_attrs(
    parent: xml.dom.minidom.Element,
    namespace: str,
    local_name: str,
    attr_names: Iterable[str],
) -> Tuple[Optional[str], List[Optional[str]]]:
    content: Optional[str]
    att_vals: List[Optional[str]]
    node_list = get_children_by_tag_name(parent, namespace, local_name)
    if not node_list:
        content = None
        att_vals = []
    elif len(node_list) == 1:
        content = get_text(node_list[0].childNodes, local_name != "code")
        att_vals = []
        for att in attr_names:
            att_vals.append(get_attribute_text(node_list[0], att))
    else:
        raise EyegradeException("Duplicate element: " + local_name)
    return content, att_vals


def get_attribute_text(
    element: xml.dom.minidom.Element, attribute_name: str
) -> Optional[str]:
    value = element.getAttributeNS(EYEGRADE_NAMESPACE, attribute_name)
    if value != "":
        return text_norm_re.sub(" ", value.strip())
    return None


def get_children_by_tag_name(
    parent: xml.dom.minidom.Element, namespace: str, local_name: str
) -> List[xml.dom.minidom.Element]:
    return get_children_by_tag_names(parent, namespace, [local_name])


def get_children_by_tag_names(
    parent: xml.dom.minidom.Element, namespace: str, local_names: Iterable[str]
) -> List[xml.dom.minidom.Element]:
    return [
        e
        for e in parent.childNodes
        if (
            e.nodeType == e.ELEMENT_NODE
            and e.localName in local_names
            and e.namespaceURI == namespace
        )
    ]


def get_text(
    node_list: xml.dom.minicompat.NodeList, normalize: bool = True
) -> Optional[str]:
    data = []
    for node in node_list:
        if node.nodeType == node.TEXT_NODE:
            data.append(node.data)
    if data:
        text = "".join(data)
        if normalize:
            return text_norm_re.sub(" ", text.strip())
        return text.strip()
    return None


def get_full_name(
    element: xml.dom.minidom.Element,
) -> Tuple[Optional[str], Optional[str]]:
    """Returns a tuple with (namespace, local_name) for the given element."""
    return (element.namespaceURI, element.localName)


def printable_name(element: xml.dom.minidom.Element) -> str:
    """Returns a string 'namespace:local_name' for the given element."""
    return "{}:{}".format(element.namespaceURI, element.localName)



# Eyegrade: grading multiple choice questions with a webcam
# Copyright (C) 2010-2021 Jesus Arias Fisteus
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see
# <https://www.gnu.org/licenses/>.
#

import os
import re
import configparser

from typing import Union

from . import utils
from . import scoring
from . import students


utils.EyegradeException.register_error(
    "exam-config-parse-error",
    "A parsing error occurred in the exam configuration file.",
)


class Exam:
    def __init__(
        self,
        capture_,
        decisions,
        solutions,
        student_listings,
        exam_id,
        question_scores,
        sessiondb=None,
    ):
        self.capture = capture_
        self.decisions = decisions
        self.student_listings = student_listings
        self.exam_id = exam_id
        self.score = scoring.Score(decisions.answers, solutions, question_scores)
        rank = self.rank_students()
        self.decisions.set_students_rank(rank)
        if rank:
            self.decisions.set_student(rank[0])
        self.sessiondb = sessiondb

    def update_question_scores(self, question_scores):
        self.score.update_question_scores(question_scores)

    def reset_image(self):
        self.capture.reset_image()

    def draw_answers(self):
        self.capture.draw_answers(self.score)

    def draw_status(self):
        self.capture.draw_status()

    def draw_corner(self, point):
        self.capture.draw_corner(point)

    def get_image_drawn(self):
        return self.capture.image_drawn

    def toggle_answer(self, question, answer):
        if self.decisions.answers[question] == answer:
            self.decisions.change_answer(question, 0)
        else:
            self.decisions.change_answer(question, answer)
        self.score.update()
        self.capture.reset_image()
        self.draw_answers()

    def rank_students(self):
        if self.decisions.detected_id is not None:
            rank = [
                (self._id_rank(s, self.decisions.id_scores), s)
                for s in self.student_listings.iter_students()
            ]
            # Only the top-5 students are passed to the rest of the program
            students_rank = [student for _, student in sorted(rank, reverse=True)[:5]]
            if not students_rank:
                students_rank = [
                    students.Student(self.decisions.detected_id, None, None, None, None)
                ]
        else:
            # Take the first student as the default one
            students_rank = [next(self.student_listings.iter_students())]
        return students_rank

    def get_student_id_and_name(self):
        if self.decisions.student is not None:
            return self.decisions.student.id_and_name
        else:
            return None

    def ranked_student_ids(self):
        """Returns the ranked list of students as taken from the decision.

        Each entry is a student object. They are ranked according to
        their probability to be the actual student id. The most probable
        is the first in the list.

        """
        if (
            self.decisions.students_rank
            and self.decisions.student is not None
            and self.decisions.students_rank[0] != self.decisions.student
        ):
            rank = list(self.decisions.students_rank)
            if self.decisions.student in rank:
                rank.remove(self.decisions.student)
            rank.insert(0, self.decisions.student)
        else:
            rank = self.decisions.students_rank
        return rank

    def update_student_id(self, student):
        """Updates the student id of the current exam.

        Receives the Student object of the new identity
        (or None for clearing the student identity).

        """
        self.decisions.set_student(student)

    def load_capture(self):
        if self.capture is None:
            self.capture = self.sessiondb.read_capture(self.exam_id)

    def clear_capture(self):
        if self.capture is not None:
            self.capture = None

    def image_drawn_path(self):
        image_name = utils.capture_name(
            self.sessiondb.exam_config.capture_pattern,
            self.exam_id,
            self.decisions.student,
        )
        path = os.path.join(self.sessiondb.session_dir, "captures", image_name)
        if not os.path.isfile(path):
            path = utils.resource_path("not_found.png")
        return path

    def _id_rank(self, student, scores):
        rank = 0.0
        if len(scores) == len(student.student_id):
            for i, digit in enumerate(student.student_id):
                rank += scores[i][int(digit)]
        return rank


class ExamConfig:
    """Class for representing exam configuration. Once an instance has
    been created and data loaded, access directly to the attributes
    to get the data. The constructor reads data from a file. See
    doc/exam-data.sample for an example of such a file."""

    SCORES_MODE_NONE = 1
    SCORES_MODE_WEIGHTS = 2
    SCORES_MODE_INDIVIDUAL = 3

    re_model = re.compile("model-[0a-zA-Z]")

    def __init__(self, filename=None, capture_pattern=None):
        """Loads data from file if 'filename' is not None. Otherwise,
        default values are assigned to the attributes."""
        if filename is not None:
            self.read(filename=filename)
        else:
            self.num_questions = 0
            self.solutions = {}
            self.id_num_digits = 0
            self.dimensions = []
            self.num_options = []
            self.permutations = {}
            self.model_variations = {}
            self.question_variations = {}
            self.models = []
            self.scores = {}
            self.base_scores = None
            self.left_to_right_numbering = False
            self.survey_mode = None
            self.scores_mode = ExamConfig.SCORES_MODE_NONE
        if capture_pattern is not None:
            self.capture_pattern = capture_pattern
        else:
            self.capture_pattern = utils.default_capture_pattern

    def __eq__(self, other):
        if isinstance(other, self.__class__):
            equal = self.as_tuple() == other.as_tuple()
        else:
            equal = False
        return equal

    def as_tuple(self):
        return (
            self.num_questions,
            self.solutions,
            self.id_num_digits,
            self.dimensions,
            self.permutations,
            self.model_variations,
            self.get_all_question_variations(),
            self.models,
            self.scores,
            self.base_scores,
            self.left_to_right_numbering,
            self.survey_mode,
            self.scores_mode,
        )

    def add_model(self, model):
        if model not in self.models:
            self.models.append(model)

    def set_solutions(self, model, solutions):
        if not isinstance(solutions, list):
            solutions = self._parse_solutions(solutions)
        if len(solutions) != self.num_questions:
            raise utils.EyegradeException(
                "Incorrect number of solutions (got {}, expected {}) for model {}.".format(
                    len(solutions), self.num_questions, model
                ),
                key="exam-config-parse-error",
            )
        for solution_set, num_options in zip(solutions, self.num_options):
            for solution in solution_set:
                if solution < 1 or solution > num_options:
                    raise utils.EyegradeException(
                        "Solution out of range for model {} (got {}, expected between 1 and {}).".format(
                            model, solution, num_options
                        ),
                        key="exam-config-parse-error",
                    )
            self.solutions[model] = solutions
        self.add_model(model)

    def get_solutions(self, model):
        """Returns the solutions for the given model.

        If in survey mode it returns []. If there are no solutions for
        this model, it returns None.

        """
        if not self.survey_mode:
            if model in self.solutions:
                return self.solutions[model]
            else:
                return None
        else:
            return []

    def set_permutations(self, model, permutations):
        if not isinstance(permutations, list):
            permutations = self._parse_permutations(permutations)
        elif permutations and isinstance(permutations[0], str):
            permutations = [
                self._parse_permutation(p, i) for i, p in enumerate(permutations)
            ]
        if len(permutations) != self.num_questions:
            raise ValueError("Permutations with incorrect number of questions")
        self.permutations[model] = permutations
        self.add_model(model)

    def get_permutations(self, model):
        """Returns the permutations for the given model.

        If there are no permutations for this model, it returns None.

        """
        if model in self.permutations:
            return self.permutations[model]
        else:
            return None

    def set_question_variations(self, model, variations):
        if not isinstance(variations, list):
            variations = self._parse_variations(variations)
        if len(variations) != self.num_questions:
            raise ValueError("Variations with incorrect number of questions")
        self.question_variations[model] = variations
        self.add_model(model)

    def get_model_variation(self, model):
        """Returns the general variation for the given model or None if not set."""
        if self.model_variations:
            return self.model_variations.get(model, None)

    def get_question_variations(self, model):
        """Returns the variations for the given model.

        If there are no variations for this model, it returns the default
        ones instead of None.

        """
        if model in self.question_variations:
            return self.question_variations[model]
        else:
            return [0] * self.num_questions

    def get_all_question_variations(self):
        """Returns variations for all the registered models.

        The difference with respect to just accessing the variations attribute
        is that, when no variations are declared at all or for some models,
        the default variations are returned instead of None.

        """
        variations = {}
        for model in self.models:
            variations[model] = self.get_question_variations(model)
        return variations

    def set_dimensions(self, dimensions):
        self.dimensions, self.num_options = utils.parse_dimensions(dimensions)
        self.num_questions = sum(dim[1] for dim in self.dimensions)

    def enter_score_mode_none(self):
        """Resets the object to no scores."""
        self.scores_mode = ExamConfig.SCORES_MODE_NONE
        self.base_scores = None
        self.scores = {}

    def set_base_scores(self, scores, same_weights=False):
        """Set the base scores for the questions of this exam.

        The `scores` parameter must be an instance of QuestionScores.
        It must be done before setting the weights of each question.

        """
        if scores.weight != 1:
            raise ValueError("The base score must have weigth 1")
        if self.scores_mode == ExamConfig.SCORES_MODE_NONE:
            self.scores_mode = ExamConfig.SCORES_MODE_WEIGHTS
        elif self.scores_mode != ExamConfig.SCORES_MODE_WEIGHTS:
            raise ValueError("The score mode does not allow base scores")
        self.base_scores = scores
        if same_weights:
            self.reset_question_weights()
            for model in self.models:
                self.set_equal_scores(model)

    def set_equal_scores(self, model):
        """Set the base scores for the questions of this exam.

        The `scores` parameter must be an instance of QuestionScores.
        It must be done before setting the weights of each question.

        """
        if (
            self.scores_mode != ExamConfig.SCORES_MODE_WEIGHTS
            or self.base_scores is None
        ):
            raise ValueError("Invalid scores mode for set_equal_scores")
        scores = [
            self.base_scores.clone(new_weight=1) for i in range(self.num_questions)
        ]
        self._set_question_scores_internal(model, scores)

    def set_question_weights(self, model, weights):
        """Set the scores for a given model from question weights.

        The `weights` parameter can be a list with the weight of each
        question of that model.
        The final scores must be the same as in the scores for the rest of
        the models, possibly in a different order.

        A base score must have already been set.

        """
        if self.scores_mode != ExamConfig.SCORES_MODE_WEIGHTS:
            raise ValueError("Not in scores weight mode.")
        if isinstance(weights, str):
            weights = self._parse_weights(weights)
        scores = [self.base_scores.clone(new_weight=weight) for weight in weights]
        self._set_question_scores_internal(model, scores)

    def get_question_weights(self, model, formatted=False):
        """Return the list of question weights for a given model.

        Returns None if scores are not set by means of a base score,
        or there are no scores for this model.

        """
        if (
            self.scores_mode != ExamConfig.SCORES_MODE_WEIGHTS
            or self.base_scores is None
            or model not in self.scores
        ):
            return None
        elif not formatted:
            return [s.weight for s in self.scores[model]]
        else:
            return [s.format_weight() for s in self.scores[model]]

    def reset_question_weights(self):
        self.scores = {}

    def all_weights_are_one(self):
        """Return True if all the score weights are 1.

        It returns False if there are no scores set for at least one model.

        """
        if self.scores:
            # We only need to check one list of scores
            return all(s.weight == 1 for s in next(iter(self.scores.values())))
        else:
            return False

    def set_question_scores(self, model, scores):
        """Set the scores for a given model from question weights.

        The `scores` parameter must be a list of QuestionScores objects.
        The scores must be the same as in the scores for the rest of
        the models, possibly in a different order.

        A base score cannot have already been set.

        """
        if self.scores_mode == ExamConfig.SCORES_MODE_NONE:
            self.scores_mode = ExamConfig.SCORES_MODE_INDIVIDUAL
        elif self.scores_mode != ExamConfig.SCORES_MODE_INDIVIDUAL:
            raise ValueError("Invalid scores mode at set_question_scores")
        for score in scores:
            if score.weight != 1:
                raise ValueError("Only weight 1 scores allowed")
        self._set_question_scores_internal(model, scores)

    def _set_question_scores_internal(self, model, scores):
        """Set the scores for a given model from question weights.

        Internal method that does not check the current mode.

        The `scores` parameter must be a list of QuestionScores objects.
        The scores must be the same as in the scores for the rest of
        the models, possibly in a different order.

        """
        if len(scores) != self.num_questions:
            raise ValueError("Scores with an incorrect number of questions")
        if self.scores and sorted(scores) != sorted(next(iter(self.scores.values()))):
            #        if self.scores and sorted(scores) != sorted(self.scores.values()[0]):
            raise ValueError("Scores for all models must be equal but their order")
        self.scores[model] = scores
        self.add_model(model)

    def get_num_choices(self):
        """Returns the number of choices per question.

        If not all the questions have the same number of choices, it returns
        the maximum number of choices. If there are no questions, it returns
        None.

        """
        choices = [dim[0] for dim in self.dimensions]
        if choices:
            return max(choices)
        else:
            return None

    def read(self, filename):
        exam_data = configparser.ConfigParser()
        files_read = exam_data.read([filename])
        if len(files_read) != 1:
            raise IOError("Exam config file not found: " + filename)
        self._read_config_parser(exam_data)

    def _read_config_parser(self, exam_data: configparser.ConfigParser):
        """Reads exam configuration."""
        try:
            self.id_num_digits = exam_data.getint("exam", "id-num-digits")
        except ValueError:
            self.id_num_digits = 0
        self.set_dimensions(exam_data.get("exam", "dimensions"))
        has_solutions = exam_data.has_section("solutions")
        has_permutations = exam_data.has_section("permutations")
        has_model_variations = exam_data.has_section("model-variations")
        has_question_variations = exam_data.has_section("variations")
        self.solutions = {}
        self.permutations = {}
        self.model_variations = {}
        self.question_variations = {}
        self.models = []
        if has_solutions:
            for key, value in exam_data.items("solutions"):
                if not self.re_model.match(key):
                    raise utils.EyegradeException(
                        "Incorrect key in exam config: " + key,
                        key="exam-config-parse-error",
                    )
                model = key[-1].upper()
                self.set_solutions(model, value)
                if has_permutations:
                    key = "permutations-" + model
                    value = exam_data.get("permutations", key)
                    self.set_permutations(model, value)
                if has_model_variations:
                    key = "model-variation-" + model
                    value = exam_data.get("model-variations", key)
                    self.model_variations[model] = int(value)
                if has_question_variations:
                    key = "variations-" + model
                    value = exam_data.get("variations", key)
                    self.set_question_variations(model, value)
        has_correct_weight = exam_data.has_option("exam", "correct-weight")
        has_incorrect_weight = exam_data.has_option("exam", "incorrect-weight")
        has_blank_weight = exam_data.has_option("exam", "blank-weight")
        self.scores = {}
        if has_correct_weight and has_incorrect_weight:
            correct_weight = exam_data.get("exam", "correct-weight")
            incorrect_weight = exam_data.get("exam", "incorrect-weight")
            blank_weight: Union[str, int]
            if has_blank_weight:
                blank_weight = exam_data.get("exam", "blank-weight")
            else:
                blank_weight = 0
            self.scores_mode = ExamConfig.SCORES_MODE_WEIGHTS
            base_scores = scoring.QuestionScores(
                correct_weight, incorrect_weight, blank_weight
            )
            if not exam_data.has_section("question-score-weights"):
                self.set_base_scores(base_scores, same_weights=True)
            else:
                self.set_base_scores(base_scores)
                for model in self.models:
                    key = "weights-" + model
                    value = exam_data.get("question-score-weights", key)
                    self.set_question_weights(model, value)
        elif not has_correct_weight and not has_incorrect_weight:
            self.base_scores = None
            self.scores_mode = ExamConfig.SCORES_MODE_NONE
        else:
            raise utils.EyegradeException(
                "Exam config must contain correct and incorrect weight or none.",
                key="exam-config-parse-error",
            )
        if exam_data.has_option("exam", "left-to-right-numbering"):
            self.left_to_right_numbering = exam_data.getboolean(
                "exam", "left-to-right-numbering"
            )
        else:
            self.left_to_right_numbering = False
        if exam_data.has_option("exam", "survey-mode"):
            self.survey_mode = exam_data.getboolean("exam", "survey-mode")
        else:
            self.survey_mode = False
        self.models.sort()

    def save(self, filename):
        data = []
        data.append("[exam]")
        data.append("dimensions: %s" % self.format_dimensions())
        data.append("id-num-digits: %d" % self.id_num_digits)
        if self.left_to_right_numbering:
            data.append("left-to-right-numbering: yes")
        if self.survey_mode:
            data.append("survey-mode: yes")
        if self.base_scores is not None:
            data.append(
                "correct-weight: {0}".format(self.base_scores.format_correct_score())
            )
            data.append(
                "incorrect-weight: {0}".format(
                    self.base_scores.format_incorrect_score()
                )
            )
            data.append(
                "blank-weight: {0}".format(self.base_scores.format_blank_score())
            )
        if self.solutions:
            data.append("")
            data.append("[solutions]")
            for model in sorted(self.models):
                data.append(
                    "model-{0}: {1}".format(model, self.format_solutions(model))
                )
        if self.permutations:
            data.append("")
            data.append("[permutations]")
            for model in sorted(self.models):
                data.append(
                    "permutations-{0}: {1}".format(
                        model, self.format_permutations(model)
                    )
                )
        if self.model_variations:
            data.append("")
            data.append("[model-variations]")
            for model in sorted(self.models):
                data.append(
                    "model-variation-{0}: {1}".format(
                        model, self.model_variations.get(model, 0)
                    )
                )
        if self.question_variations:
            data.append("")
            data.append("[variations]")
            for model in sorted(self.models):
                data.append(
                    "variations-{0}: {1}".format(model, self.format_variations(model))
                )
        if (
            self.scores_mode == ExamConfig.SCORES_MODE_WEIGHTS
            and self.scores
            and not self.all_weights_are_one()
        ):
            # If all the scores are equal, there is no need to specify weights
            data.append("")
            data.append("[question-score-weights]")
            for model in sorted(self.models):
                data.append(
                    "weights-{0}: {1}".format(model, self.format_weights(model))
                )
        data.append("")
        with open(filename, "w") as file_:
            file_.write("\n".join(data))

    def format_dimensions(self):
        return ";".join(["%d,%d" % (cols, rows) for cols, rows in self.dimensions])

    def format_solutions(self, model):
        return "/".join(
            [self._format_question_solutions(n) for n in self.solutions[model]]
        )

    def _format_question_solutions(self, question_solution):
        return ",".join(str(s) for s in question_solution)

    def format_permutations(self, model):
        return "/".join([self.format_permutation(p) for p in self.permutations[model]])

    def format_permutation(self, permutation):
        num_question, options = permutation
        return "%d{%s}" % (num_question, ",".join([str(n) for n in options]))

    def format_variations(self, model):
        return "/".join(
            [str(variation) for variation in self.get_question_variations(model)]
        )

    def format_weights(self, model):
        return ",".join([s.format_weight() for s in self.scores[model]])

    def _parse_solutions(self, solutions_str):
        pieces = solutions_str.split("/")
        if len(pieces) != self.num_questions:
            raise utils.EyegradeException(
                "Wrong number of solutions", key="exam-config-parse-error"
            )
        return [self._parse_question_solution(piece) for piece in pieces]

    def _parse_question_solution(self, text):
        pieces = text.split(",")
        if not pieces:
            raise utils.EyegradeException(
                "Wrong number of solutions to a question", key="exam-config-parse-error"
            )
        return set(int(p) for p in pieces)

    def _parse_permutations(self, permutations_str):
        permutations = []
        pieces = permutations_str.split("/")
        if len(pieces) != self.num_questions:
            raise utils.EyegradeException(
                "Wrong number of permutations", key="exam-config-parse-error"
            )
        for i, piece in enumerate(pieces):
            permutations.append(self._parse_permutation(piece, i))
        return permutations

    def _parse_permutation(self, str_value, question_number):
        splitted = str_value.split("{")
        num_question = int(splitted[0])
        options = [int(p) for p in splitted[1][:-1].split(",")]
        if len(options) > self.num_options[question_number]:
            raise utils.EyegradeException(
                "Wrong number of options in permutation", key="exam-config-parse-error"
            )
        return (num_question, options)

    def _parse_variations(self, variations_str):
        pieces = variations_str.split("/")
        if len(pieces) != self.num_questions:
            raise utils.EyegradeException(
                "Wrong number of variation items", key="exam-config-parse-error"
            )
        return [int(variation) for variation in pieces]

    def _parse_weights(self, weights_str):
        pieces = weights_str.split(",")
        if len(pieces) != self.num_questions:
            raise utils.EyegradeException(
                "Wrong number of weight items", key="exam-config-parse-error"
            )
        return [p for p in pieces]
